@@ -1,4 +1,5 @@
 ### Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2009-2010 Kai Willadsen <kai.willadsen@gmail.com>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -14,7 +15,9 @@
 ### along with this program; if not, write to the Free Software
 ### Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-import filecmp
+import collections
+import copy
+import errno
 import paths
 from ui import gnomeglade
 import gtk
@@ -33,6 +36,9 @@ import time
 
 import ui.emblemcellrenderer
 
+from util.namedtuple import namedtuple
+from meldapp import app
+
 gdk = gtk.gdk
 
 ################################################################################
@@ -41,65 +47,118 @@ gdk = gtk.gdk
 #
 ################################################################################
 
+# For compatibility with Python 2.5, we use the Python 2.4 compatible version of
+# namedtuple. The class is included in the collections module as of Python 2.6.
+class StatItem(namedtuple('StatItem', 'mode size time')):
+    __slots__ = ()
+
+    @classmethod
+    def _make(cls, stat_result):
+        return StatItem(stat.S_IFMT(stat_result.st_mode),
+                        stat_result.st_size, stat_result.st_mtime)
+
+
+CacheResult = namedtuple('CacheResult', 'stats result')
+
+
 _cache = {}
+Same, SameFiltered, DodgySame, DodgyDifferent, Different, FileError = range(6)
+# TODO: Get the block size from os.stat
+CHUNK_SIZE = 4096
 
-def _files_same(lof, regexes):
-    """Return 1 if all the files in 'lof' have the same contents.
-       If the files are the same after the regular expression substitution, return 2.
-       Finally, return 0 if the files still differ.
+
+def all_same(lst):
+    return not lst or lst.count(lst[0]) == len(lst)
+
+
+def _files_same(files, regexes):
+    """Determine whether a list of files are the same.
+
+    Possible results are:
+      Same: The files are the same
+      SameFiltered: The files are identical only after filtering with 'regexes'
+      DodgySame: The files are superficially the same (i.e., type, size, mtime)
+      DodgyDifferent: The files are superficially different
+      FileError: There was a problem reading one or more of the files
     """
-    # early out if only one file
-    if len(lof) <= 1:
-        return 1
-    # get sigs
-    lof = tuple(lof)
-    def sig(f):
-        s = os.stat(f)
-        return misc.struct(mode=stat.S_IFMT(s.st_mode), size=s.st_size, time=s.st_mtime)
-    def all_same(l):
-        for i in l[1:]:
-            if l[0] != i:
-                return 0
-        return 1
-    sigs = tuple( [ sig(f) for f in lof ] )
-    # check for directories
-    arefiles = [ stat.S_ISREG(s.mode) for s in sigs ]
-    if arefiles.count(0) == len(arefiles): # all dirs
-        return 1
-    elif arefiles.count(0): # mixture
-        return 0
-    # if no substitutions look for different sizes
-    if len(regexes) == 0 and all_same( [s.size for s in sigs] ) == 0:
-        return 0
-    # try cache
-    try:
-        cache = _cache[ lof ]
-    except KeyError:
-        pass
-    else:
-        if cache.sigs == sigs: # up to date
-            return cache.result
-    # do it
-    try:
-        contents = [open(f, "r").read() for f in lof]
-    except (MemoryError, OverflowError): # Files are too large
-        # FIXME: Filters are not current applied in this case. If that was
-        # to be fixed, we could drop the all-at-once loading.
-        for i in range(len(lof) - 1):
-            same = filecmp.cmp(lof[i], lof[i + 1], False)
-            if not same:
-                return 0
-        return 1
 
-    if all_same(contents):
-        result = 1
-    else:
+    # One file is the same as itself
+    if len(files) < 2:
+        return Same
+
+    files = tuple(files)
+    regexes = tuple(regexes)
+    stats = tuple([StatItem._make(os.stat(f)) for f in files])
+
+    # If all entries are directories, they are considered to be the same
+    if all([stat.S_ISDIR(s.mode) for s in stats]):
+        return Same
+
+    # If any entries are not regular files, consider them different
+    if not all([stat.S_ISREG(s.mode) for s in stats]):
+        return Different
+
+    # If there are no text filters, unequal sizes imply a difference
+    if not regexes and not all_same([s.size for s in stats]):
+        return Different
+
+    # Check the cache before doing the expensive comparison
+    cache = _cache.get((files, regexes))
+    if cache and cache.stats == stats:
+        return cache.result
+
+    # Open files and compare bit-by-bit
+    contents = [[] for f in files]
+    result = None
+
+    try:
+        handles = [open(f, "rb") for f in files]
+        try:
+            data = [h.read(CHUNK_SIZE) for h in handles]
+
+            # Rough test to see whether files are binary. If files are guessed
+            # to be binary, we unset regexes for speed and space reasons.
+            if any(["\0" in d for d in data]):
+                regexes = tuple()
+
+            while True:
+                if all_same(data):
+                    if not data[0]:
+                        break
+                else:
+                    result = Different
+                    if not regexes:
+                        break
+
+                if regexes:
+                    for i in range(len(data)):
+                        contents[i].append(data[i])
+
+                data = [h.read(CHUNK_SIZE) for h in handles]
+
+        # Files are too large; we can't apply filters
+        except (MemoryError, OverflowError):
+            result = DodgySame if all_same(stats) else DodgyDifferent
+        finally:
+            for h in handles:
+                h.close()
+    except IOError:
+        # Don't cache generic errors as results
+        return FileError
+
+    if result is None:
+        result = Same
+
+    if result == Different and regexes:
+        contents = ["".join(c) for c in contents]
         for r in regexes:
-            contents = [ re.sub(r, "", c) for c in contents ]
-        result = all_same(contents) and 2
-    _cache[ lof ] = misc.struct(sigs=sigs, result=result)
+            contents = [re.sub(r, "", c) for c in contents]
+        result = SameFiltered if all_same(contents) else Different
+
+    _cache[(files, regexes)] = CacheResult(stats, result)
     return result
 
+
 COL_EMBLEM, COL_END = tree.COL_END, tree.COL_END + 1
 
 ################################################################################
@@ -112,18 +171,32 @@ class DirDiffTreeStore(tree.DiffTreeStore):
         types = [str] * COL_END * ntree
         tree.DiffTreeStore.__init__(self, ntree, types)
 
-################################################################################
-#
-# TypeFilter
-#
-################################################################################
 
-class TypeFilter(object):
-    __slots__ = ("label", "filter", "active")
-    def __init__(self, label, active, filter):
-        self.label = label
-        self.active = active
-        self.filter = filter
+class CanonicalListing(object):
+    """Multi-pane lists with canonicalised matching and error detection"""
+
+    def __init__(self, n, canonicalize=None):
+        self.items = collections.defaultdict(lambda: [None] * n)
+        self.errors = []
+        if canonicalize is not None:
+            self.canonicalize = canonicalize
+            self.add = self.add_canon
+
+    def add(self, pane, item):
+        self.items[item][pane] = item
+
+    def add_canon(self, pane, item):
+        ci = self.canonicalize(item)
+        if self.items[ci][pane] is None:
+            self.items[ci][pane] = item
+        else:
+            self.errors.append((pane, item, self.items[ci][pane]))
+
+    def get(self):
+        first = lambda seq: next(s for s in seq if s)
+        filled = lambda seq: tuple([s or first(seq) for s in seq])
+        return sorted([filled(v) for v in self.items.itervalues()])
+
 
 ################################################################################
 #
@@ -140,9 +213,9 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
         actions = (
             ("DirCompare",   gtk.STOCK_DIALOG_INFO,  _("_Compare"), None, _("Compare selected"), self.on_button_diff_clicked),
-            ("DirCopyLeft",  gtk.STOCK_GO_BACK,      _("Left"),     None, _("Copy To Left"), self.on_button_copy_left_clicked),
-            ("DirCopyRight", gtk.STOCK_GO_FORWARD,   _("Right"),    None, _("Copy To Right"), self.on_button_copy_right_clicked),
-            ("DirDelete",    gtk.STOCK_DELETE,        None,         None, _("Delete selected"), self.on_button_delete_clicked),
+            ("DirCopyLeft",  gtk.STOCK_GO_BACK,      _("Copy _Left"),     "<Alt>Left", _("Copy to left"), self.on_button_copy_left_clicked),
+            ("DirCopyRight", gtk.STOCK_GO_FORWARD,   _("Copy _Right"),    "<Alt>Right", _("Copy to right"), self.on_button_copy_right_clicked),
+            ("DirDelete",    gtk.STOCK_DELETE,        None,         "Delete", _("Delete selected"), self.on_button_delete_clicked),
             ("Hide",         gtk.STOCK_NO,           _("Hide"),     None, _("Hide selected"), self.on_filter_hide_current_clicked),
 
             ("DirOpen",      gtk.STOCK_OPEN,          None,         None, _("Open selected"), self.on_button_open_clicked),
@@ -161,12 +234,19 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.actiongroup.set_translation_domain("meld")
         self.actiongroup.add_actions(actions)
         self.actiongroup.add_toggle_actions(toggleactions)
+        self.name_filters = []
         self.create_name_filters()
+        app.connect("file-filters-changed", self.on_file_filters_changed)
+        self.text_filters = []
+        self.create_text_filters()
+        app.connect("text-filters-changed", self.on_text_filters_changed)
         for button in ("DirCompare", "DirCopyLeft", "DirCopyRight",
                        "DirDelete", "Hide", "IgnoreCase", "ShowSame",
                        "ShowNew", "ShowModified", "CustomFilterMenu"):
             self.actiongroup.get_action(button).props.is_important = True
-        self.map_widgets_into_lists( ["treeview", "fileentry", "diffmap", "scrolledwindow", "linkmap"] )
+        self.map_widgets_into_lists(["treeview", "fileentry", "scrolledwindow",
+                                     "diffmap", "linkmap", "msgarea_mgr",
+                                     "vbox"])
         self.set_num_panes(num_panes)
         self.focus_in_events = []
         self.focus_out_events = []
@@ -200,17 +280,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             tree.STATE_MODIFIED,
             tree.STATE_NEW,
         ]
-        self.update_regexes()
-
-    def update_regexes(self):
-        self.regexes = []
-        for r in [ misc.ListItem(i) for i in self.prefs.regexes.split("\n") ]:
-            if r.active:
-                try:
-                    self.regexes.append( re.compile(r.value+"(?m)") )
-                except re.error:
-                    misc.run_dialog(
-                        text=_("Error converting pattern '%s' to regular expression") % r.value )
 
     def _custom_popup_deactivated(self, popup):
         self.filter_menu_button.set_active(False)
@@ -221,8 +290,15 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.custom_popup.popup(None, None, misc.position_menu_under_widget,
                                     1, gtk.get_current_event_time(), self.filter_menu_button)
 
-    def on_container_switch_in_event(self, ui):
-        melddoc.MeldDoc.on_container_switch_in_event(self, ui)
+    def _cleanup_filter_menu_button(self, ui):
+        if self.popup_deactivate_id:
+            self.popup_menu.disconnect(self.popup_deactivate_id)
+        if self.custom_merge_id:
+            ui.remove_ui(self.custom_merge_id)
+        if self.filter_actiongroup in ui.get_action_groups():
+            ui.remove_action_group(self.filter_actiongroup)
+
+    def _create_filter_menu_button(self, ui):
         ui.insert_action_group(self.filter_actiongroup, -1)
         self.custom_merge_id = ui.new_merge_id()
         for x in self.filter_ui:
@@ -233,50 +309,68 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         label = misc.make_tool_button_widget(self.filter_menu_button.props.label)
         self.filter_menu_button.set_label_widget(label)
 
+    def on_container_switch_in_event(self, ui):
+        melddoc.MeldDoc.on_container_switch_in_event(self, ui)
+        self._create_filter_menu_button(ui)
+        self.ui_manager = ui
+
+        # FIXME: Add real sensitivity handling
+        self.emit("next-diff-changed", True, True)
         if self.treeview_focussed:
             self.scheduler.add_task(self.treeview_focussed.grab_focus)
             self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def on_container_switch_out_event(self, ui):
-        self.popup_menu.disconnect(self.popup_deactivate_id)
-        ui.remove_ui(self.custom_merge_id)
-        ui.remove_action_group(self.filter_actiongroup)
+        self._cleanup_filter_menu_button(ui)
         melddoc.MeldDoc.on_container_switch_out_event(self, ui)
 
-    def create_name_filters(self):
-        self.name_filters_available = []
-        for f in [misc.ListItem(s) for s in self.prefs.filters.split("\n") ]:
-            bits = f.value.split()
-            if len(bits) > 1:
-                regex = "(%s)$" % "|".join( [misc.shell_to_regex(b)[:-1] for b in bits] )
-            elif len(bits):
-                regex = misc.shell_to_regex(bits[0])
-            else: # an empty pattern would match anything, skip it
-                continue
-            try:
-                cregex = re.compile(regex)
-            except re.error:
-                misc.run_dialog( _("Error converting pattern '%s' to regular expression") % f.value, self )
-            else:
-                func = lambda x, r=cregex : r.match(x) is None
-                self.name_filters_available.append( TypeFilter(f.name, f.active, func) )
-        self.name_filters = [f for f in self.name_filters_available if f.active]
+    def on_file_filters_changed(self, app):
+        self._cleanup_filter_menu_button(self.ui_manager)
+        relevant_change = self.create_name_filters()
+        self._create_filter_menu_button(self.ui_manager)
+        if relevant_change:
+            self.refresh()
 
+    def create_name_filters(self):
+        # Ordering of name filters is irrelevant
+        old_active = set([f.filter_string for f in self.name_filters if f.active])
+        new_active = set([f.filter_string for f in app.file_filters if f.active])
+        active_filters_changed = old_active != new_active
+
+        self.name_filters = [copy.copy(f) for f in app.file_filters]
         actions = []
+        disabled_actions = []
         self.filter_ui = []
-        for i,f in enumerate(self.name_filters_available):
+        for i, f in enumerate(self.name_filters):
             name = "Hide%d" % i
             callback = lambda b, i=i: self._update_name_filter(b, i)
             actions.append((name, None, f.label, None, _("Hide %s") % f.label, callback, f.active))
             self.filter_ui.append(["/CustomPopup" , name, name, gtk.UI_MANAGER_MENUITEM, False])
             self.filter_ui.append(["/Menubar/ViewMenu/FileFilters" , name, name, gtk.UI_MANAGER_MENUITEM, False])
+            if f.filter is None:
+                disabled_actions.append(name)
 
         self.filter_actiongroup = gtk.ActionGroup("DirdiffFilterActions")
         self.filter_actiongroup.add_toggle_actions(actions)
+        for name in disabled_actions:
+            self.filter_actiongroup.get_action(name).set_sensitive(False)
 
-    def on_preference_changed(self, key, value):
-        if key == "regexes":
-            self.update_regexes()
+        return active_filters_changed
+
+    def on_text_filters_changed(self, app):
+        relevant_change = self.create_text_filters()
+        if relevant_change:
+            self.refresh()
+
+    def create_text_filters(self):
+        # In contrast to file filters, ordering of text filters can matter
+        old_active = [f.filter_string for f in self.text_filters if f.active]
+        new_active = [f.filter_string for f in app.text_filters if f.active]
+        active_filters_changed = old_active != new_active
+
+        self.text_filters = [copy.copy(f) for f in app.text_filters]
+
+        return active_filters_changed
 
     def _do_to_others(self, master, objects, methodname, args):
         if not hasattr(self, "do_to_others_lock"):
@@ -354,129 +448,170 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.actiongroup.get_action("Hide").set_sensitive(False)
         yield _("[%s] Scanning %s") % (self.label_text, "")
         prefixlen = 1 + len( self.model.value_path( self.model.get_iter(rootpath), 0 ) )
-        symlinks_followed = {} # only follow symlinks once
+        symlinks_followed = set()
         todo = [ rootpath ]
-        expanded = {}
+        expanded = set()
+
+        shadowed_entries = []
+        invalid_filenames = []
         while len(todo):
             todo.sort() # depth first
             path = todo.pop(0)
             it = self.model.get_iter( path )
             roots = self.model.value_paths( it )
             yield _("[%s] Scanning %s") % (self.label_text, roots[0][prefixlen:])
-            differences = [0]
-            if not self.actiongroup.get_action("IgnoreCase").get_active():
-                class accum(object):
-                    def __init__(self, parent, roots):
-                        self.items = []
-                        self.n = parent.num_panes
-                    def add(self, pane, items):
-                        self.items.extend(items)
-                    def get(self):
-                        return [(i,) * self.n for i in sorted(set(self.items))]
-            else:
-                canonicalize = lambda x : x.lower()
-                class accum(object):
-                    def __init__(self, parent, roots):
-                        self.items = {} # map canonical names to realnames
-                        self.bad = []
-                        self.parent = parent
-                        self.roots = roots
-                        self.default = [None] * self.parent.num_panes
-                    def add(self, pane, items):
-                        for i in items:
-                            ci = canonicalize(i)
-                            try:
-                                assert self.items[ci][pane] is None
-                            except KeyError:
-                                self.items[ ci ] = self.default[:]
-                                self.items[ ci ][pane] = i
-                            except AssertionError:
-                                self.bad.append( _("'%s' hidden by '%s'") %
-                                    ( os.path.join(self.roots[pane], i), self.items[ ci ][pane]) )
-                            else:
-                                self.items[ ci ][pane] = i
-                    def get(self):
-                        if len(self.bad):
-                            misc.run_dialog(_("You are running a case insensitive comparison on"
-                                " a case sensitive filesystem. Some files are not visible:\n%s")
-                                % "\n".join( self.bad ), self.parent )
-                        keys = self.items.keys()
-                        keys.sort()
-                        def fixup(tuples): # replace None with a usable label
-                            def first_nonempty(seq):
-                                for s in seq:
-                                    if s: return s
-                            return tuple([t or first_nonempty(tuples) for t in tuples])
-                        return [fixup(self.items[k]) for k in keys]
-            accumdirs = accum(self, roots)
-            accumfiles = accum(self, roots)
-            for pane, root in enumerate(roots):
-                if os.path.isdir( root ):
-                    try:
-                        entries = os.listdir( root )
-                    except OSError, err:
-                        self.model.add_error( it, err.strerror, pane )
-                        differences = [1]
-                    else:
-                        for f in self.name_filters:
-                            entries = filter(f.filter, entries)
-                        files = []
-                        dirs = []
-                        for e in entries:
-                            try: # Necessary for some broken symlink cases; see bgo#585895
-                                s = os.lstat(os.path.join(root, e))
-                            except OSError, err:
-                                print "Ignoring OS error: %s" % err
-                                continue
-                            if stat.S_ISLNK(s.st_mode):
-                                if not self.prefs.ignore_symlinks:
-                                    key = (s.st_dev, s.st_ino)
-                                    if symlinks_followed.get( key, 0 ) == 0:
-                                        symlinks_followed[key] = 1
-                                        try:
-                                            s = os.stat(os.path.join(root, e))
-                                        except OSError, err:
-                                            print "ignoring dangling symlink", e
-                                        else:
-                                            if stat.S_ISREG(s.st_mode):
-                                                files.append(e)
-                                            elif stat.S_ISDIR(s.st_mode):
-                                                dirs.append(e)
-                            elif stat.S_ISREG(s.st_mode):
-                                files.append(e)
-                            elif stat.S_ISDIR(s.st_mode):
-                                dirs.append(e)
-                        accumfiles.add( pane, files )
-                        accumdirs.add( pane, dirs )
+            differences = False
+            encoding_errors = []
 
-            alldirs = accumdirs.get()
-            allfiles = self._filter_on_state( roots, accumfiles.get() )
+            canonicalize = None
+            if self.actiongroup.get_action("IgnoreCase").get_active():
+                canonicalize = lambda x : x.lower()
+            dirs = CanonicalListing(self.num_panes, canonicalize)
+            files = CanonicalListing(self.num_panes, canonicalize)
+
+            for pane, root in enumerate(roots):
+                if not os.path.isdir(root):
+                    continue
+
+                try:
+                    entries = os.listdir(root)
+                except OSError, err:
+                    self.model.add_error(it, err.strerror, pane)
+                    differences = True
+                    continue
+
+                for f in self.name_filters:
+                    if not f.active or f.filter is None:
+                        continue
+                    entries = [e for e in entries if f.filter.match(e) is None]
+
+                for e in entries:
+                    try:
+                        e = e.decode('utf8')
+                    except UnicodeDecodeError, err:
+                        approximate_name = e.decode('utf8', 'replace')
+                        encoding_errors.append((pane, approximate_name))
+                        continue
+
+                    try:
+                        s = os.lstat(os.path.join(root, e))
+                    # Covers certain unreadable symlink cases; see bgo#585895
+                    except OSError, err:
+                        error_string = e + err.strerror
+                        self.model.add_error(it, error_string, pane)
+                        continue
+
+                    if stat.S_ISLNK(s.st_mode):
+                        if self.prefs.ignore_symlinks:
+                            continue
+                        key = (s.st_dev, s.st_ino)
+                        if key in symlinks_followed:
+                            continue
+                        symlinks_followed.add(key)
+                        try:
+                            s = os.stat(os.path.join(root, e))
+                            if stat.S_ISREG(s.st_mode):
+                                files.add(pane, e)
+                            elif stat.S_ISDIR(s.st_mode):
+                                dirs.add(pane, e)
+                        except OSError, err:
+                            if err.errno == errno.ENOENT:
+                                error_string = e + ": Dangling symlink"
+                            else:
+                                error_string = e + err.strerror
+                            self.model.add_error(it, error_string, pane)
+                            differences = True
+                    elif stat.S_ISREG(s.st_mode):
+                        files.add(pane, e)
+                    elif stat.S_ISDIR(s.st_mode):
+                        dirs.add(pane, e)
+                    else:
+                        # FIXME: Unhandled stat type
+                        pass
+
+            for pane, f in encoding_errors:
+                invalid_filenames.append((pane, roots[pane], f))
+
+            for pane, f1, f2 in dirs.errors + files.errors:
+                shadowed_entries.append((pane, roots[pane], f1, f2))
+
+            alldirs = dirs.get()
+            allfiles = self._filter_on_state(roots, files.get())
 
             # then directories and files
             if len(alldirs) + len(allfiles) != 0:
-                def add_entry(names):
-                    child = self.model.add_entries(it, [os.path.join(r, n) for r, n in zip(roots, names)])
-                    differences[0] |= self._update_item_state(child)
-                    return child
-                map(lambda x : todo.append( self.model.get_path(add_entry(x))), alldirs )
-                map(add_entry, allfiles)
+                for names in alldirs:
+                    entries = [os.path.join(r, n) for r, n in zip(roots, names)]
+                    child = self.model.add_entries(it, entries)
+                    differences |= self._update_item_state(child)
+                    todo.append(self.model.get_path(child))
+                for names in allfiles:
+                    entries = [os.path.join(r, n) for r, n in zip(roots, names)]
+                    child = self.model.add_entries(it, entries)
+                    differences |= self._update_item_state(child)
             else: # directory is empty, add a placeholder
                 self.model.add_empty(it)
-            if differences[0]:
-                expanded[path] = False
-        for path in sorted(expanded.keys()):
-            start = path[:]
-            while len(start) and not expanded.get(start,False):
-                start = start[:-1]
-            level = len(start)
-            while level < len(path):
-                level += 1
-                cur = path[:level]
-                expanded[cur] = True
-                self.treeview[0].expand_row( cur, 0)
+            if differences:
+                expanded.add(path)
+
+        self._show_tree_wide_errors(invalid_filenames, shadowed_entries)
+
+        for path in sorted(expanded):
+            self.treeview[0].expand_to_path(path)
         yield _("[%s] Done") % self.label_text
         self.actiongroup.get_action("Hide").set_sensitive(True)
 
+    def _show_tree_wide_errors(self, invalid_filenames, shadowed_entries):
+        header = _("Multiple errors occurred while scanning this folder")
+        invalid_header = _("Files with invalid encodings found")
+        # TRANSLATORS: This is followed by a list of files
+        invalid_secondary = _("Some files were in an incorrect encoding. "
+                              "The names are something like:")
+        shadowed_header = _("Files hidden by case insensitive comparison")
+        # TRANSLATORS: This is followed by a list of files
+        shadowed_secondary = _("You are running a case insensitive comparison "
+                               "on a case sensitive filesystem. The following "
+                               "files in this folder are hidden:")
+
+        invalid_entries = [[] for i in range(self.num_panes)]
+        for pane, root, f in invalid_filenames:
+            invalid_entries[pane].append(os.path.join(root, f))
+
+        formatted_entries = [[] for i in range(self.num_panes)]
+        for pane, root, f1, f2 in shadowed_entries:
+            paths = [os.path.join(root, f) for f in (f1, f2)]
+            entry_str = _("'%s' hidden by '%s'") % (paths[0], paths[1])
+            formatted_entries[pane].append(entry_str)
+
+        if invalid_filenames or shadowed_entries:
+            for pane in range(self.num_panes):
+                invalid = "\n".join(invalid_entries[pane])
+                shadowed = "\n".join(formatted_entries[pane])
+                if invalid and shadowed:
+                    messages = (invalid_secondary, invalid, "",
+                                shadowed_secondary, shadowed)
+                elif invalid:
+                    header = invalid_header
+                    messages = (invalid_secondary, invalid)
+                elif shadowed:
+                    header = shadowed_header
+                    messages = (shadowed_secondary, shadowed)
+                else:
+                    continue
+                secondary = "\n".join(messages)
+                self.add_dismissable_msg(pane, gtk.STOCK_DIALOG_ERROR, header,
+                                         secondary)
+
+    def add_dismissable_msg(self, pane, icon, primary, secondary):
+        msgarea = self.msgarea_mgr[pane].new_from_text_and_icon(
+                        icon, primary, secondary)
+        button = msgarea.add_stock_button_with_text(_("Hi_de"),
+                        gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
+        msgarea.connect("response",
+                        lambda *args: self.msgarea_mgr[pane].clear())
+        msgarea.show_all()
+        return msgarea
+
     def launch_comparison(self, it, pane, force=1):
         """Launch comparison at 'it'. 
            If it is a file we launch a diff.
@@ -578,8 +713,9 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             fname = self.model.value_path( self.model.get_iter(paths[0]), pane )
             try:
                 stat = os.stat(fname)
-            except OSError:
-                self.emit("status-changed", "" )
+            # TypeError for if fname is None
+            except (OSError, TypeError):
+                self.emit("status-changed", "")
             else:
                 self.emit("status-changed", "%s : %s" % (rwx(stat.st_mode), nice(time.time() - stat.st_mtime) ) )
 
@@ -688,12 +824,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self._update_state_filter( tree.STATE_MODIFIED, button.get_active() )
 
     def _update_name_filter(self, button, idx):
-        for i in range(len(self.name_filters)):
-            if self.name_filters[i] == self.name_filters_available[idx]:
-                self.name_filters.pop(i)
-                break
-        if button.get_active():
-            self.name_filters.append( self.name_filters_available[idx] )
+        self.name_filters[idx].active = button.get_active()
         self.refresh()
 
     def on_filter_hide_current_clicked(self, button):
@@ -724,12 +855,13 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         """
         assert len(roots) == self.model.ntree
         ret = []
+        regexes = [f.filter for f in self.text_filters if f.active]
         for files in fileslist:
             curfiles = [ os.path.join( r, f ) for r,f in zip(roots,files) ]
             is_present = [ os.path.exists( f ) for f in curfiles ]
             all_present = 0 not in is_present
             if all_present:
-                if _files_same( curfiles, self.regexes ):
+                if _files_same(curfiles, regexes) in (Same, SameFiltered):
                     state = tree.STATE_NORMAL
                 else:
                     state = tree.STATE_MODIFIED
@@ -743,6 +875,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         """Update the state of the item at 'it'
         """
         files = self.model.value_paths(it)
+        regexes = [f.filter for f in self.text_filters if f.active]
+
         def mtime(f):
             try:
                 return os.stat(f).st_mtime
@@ -755,28 +889,33 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             newest_index = -1 # all same
         all_present = 0 not in mod_times
         if all_present:
-            all_same = _files_same( files, self.regexes )
+            all_same = _files_same(files, regexes)
             all_present_same = all_same
         else:
             lof = []
             for j in range(len(mod_times)):
                 if mod_times[j]:
                     lof.append( files[j] )
-            all_same = 0
-            all_present_same = _files_same( lof, self.regexes )
+            all_same = Different
+            all_present_same = _files_same(lof, regexes)
         different = 1
         one_isdir = [None for i in range(self.model.ntree)]
         for j in range(self.model.ntree):
             if mod_times[j]:
                 isdir = os.path.isdir( files[j] )
-                if all_same == 1:
+                # TODO: Differentiate the DodgySame case
+                if all_same == Same or all_same == DodgySame:
                     self.model.set_state(it, j,  tree.STATE_NORMAL, isdir)
                     different = 0
-                elif all_same == 2:
+                elif all_same == SameFiltered:
                     self.model.set_state(it, j,  tree.STATE_NOCHANGE, isdir)
                     different = 0
-                elif all_present_same:
+                # TODO: Differentiate the SameFiltered and DodgySame cases
+                elif all_present_same in (Same, SameFiltered, DodgySame):
                     self.model.set_state(it, j,  tree.STATE_NEW, isdir)
+                elif all_same == FileError or all_present_same == FileError:
+                    self.model.set_state(it, j,  tree.STATE_ERROR, isdir)
+                # Different and DodgyDifferent
                 else:
                     self.model.set_state(it, j,  tree.STATE_MODIFIED, isdir)
                 self.model.set_value(it,
@@ -788,13 +927,23 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.model.set_state(it, j, tree.STATE_MISSING, True in one_isdir)
         return different
 
-    def popup_in_pane(self, pane):
+    def popup_in_pane(self, pane, event):
         for (treeview, inid, outid) in zip(self.treeview, self.focus_in_events, self.focus_out_events):
             treeview.handler_block(inid)
             treeview.handler_block(outid)
         self.actiongroup.get_action("DirCopyLeft").set_sensitive(pane > 0)
         self.actiongroup.get_action("DirCopyRight").set_sensitive(pane+1 < self.num_panes)
-        self.popup_menu.popup(None, None, None, 3, gtk.get_current_event_time())
+        if event:
+            button = event.button
+            time = event.time
+        else:
+            button = 0
+            time = gtk.get_current_event_time()
+        self.popup_menu.popup(None, None, None, button, time)
+
+    def on_treeview_popup_menu(self, treeview):
+        self.popup_in_pane(self.treeview.index(treeview), None)
+        return True
 
     def on_treeview_button_press_event(self, treeview, event):
         # unselect other panes
@@ -810,7 +959,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 selected = self._get_selected_paths( self.treeview.index(treeview) )
                 if len(selected) <= 1 and event.state == 0:
                     treeview.set_cursor( path, col, 0)
-                self.popup_in_pane( self.treeview.index(treeview) )
+                self.popup_in_pane(self.treeview.index(treeview), event)
             return event.state==0
         return 0
 
@@ -821,9 +970,11 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.treeview[i].set_model(self.model)
             toshow =  self.scrolledwindow[:n] + self.fileentry[:n]
             toshow += self.linkmap[:n-1] + self.diffmap[:n]
+            toshow += self.vbox[:n] + self.msgarea_mgr[:n]
             map( lambda x: x.show(), toshow )
             tohide =  self.scrolledwindow[n:] + self.fileentry[n:]
             tohide += self.linkmap[n-1:] + self.diffmap[n:]
+            tohide += self.vbox[n:] + self.msgarea_mgr[n:]
             map( lambda x: x.hide(), tohide )
             if self.num_panes != 0: # not first time through
                 self.num_panes = n
@@ -924,7 +1075,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         for c in chunks[1:]:
             end = start + c[0]
             s,e = [int(x) for x in (math.floor(scaleit(start)), math.ceil(scaleit(end))) ]
-            gc = area.meldgc[ int(c[1]) ]
+            gc = area.meldgc[c[1]]
             if gc:
                 window.draw_rectangle( gc, 1, x0, s, x1, e-s)
                 window.draw_rectangle( area.meldgc[-1], 0, x0, s, x1, e-s)
@@ -993,7 +1144,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
         search = {gtk.gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
         for it in search( start_iter ):
-            state = int(self.model.get_state( it, pane ))
+            state = self.model.get_state(it, pane)
             if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
                 curpath = self.model.get_path(it)
                 self.treeview[pane].expand_to_path(curpath)
